from typing import Any, Type

import konfi
from konfi.source import load_fields_values
from .file import register_file_loader

__all__ = ["YAML"]


@register_file_loader(".yml", ".yaml", ".json")
class YAML(konfi.SourceABC):
    """Source which loads the config from YAML files.

    Args:
        path: File path to load file from.

        ignore_not_found: When this is set to true and the config file
            couldn't be found at the given path, no error is raised and
            nothing is loaded.
    """
    _path: str
    _ignore_not_found: bool

    def __init__(self, path: str, *, ignore_not_found: bool = False, **_) -> None:
        self._path = path
        self._ignore_not_found = ignore_not_found

    def __str__(self) -> str:
        return f"YAML: {self._path!r}"

    def load_into(self, obj: Any, template: Type) -> None:
        try:
            import yaml
        except ImportError as e:
            raise ImportError("Couldn't import 'pyyaml' package. Make sure it's installed.") from e

        try:
            with open(self._path, "r") as f:
                data = yaml.safe_load(f)
        except FileNotFoundError as e:
            if self._ignore_not_found:
                return
            else:
                raise e

        load_fields_values(obj, konfi.fields(template), data)



"""Built-in converters."""

import collections.abc as collections
import enum
import functools
import inspect
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Sequence, Set, Tuple, TypeVar, cast

from . import source, templ, typeinspect
from .converter import ComplexConverterABC, ConversionError, convert_value, has_converter, register_converter

T = TypeVar("T")


# <-- converter groups -------------------------------------------------------->

# primitive types which can be called with multiple input types to get the
# desired type.

def _register_primitive_converters(converters: Iterable[type]) -> None:
    for conv in converters:
        register_converter(conv)(conv)


_register_primitive_converters((
    bool,
    int, float, complex,
    str, bytes,
))

del _register_primitive_converters


# types which are first converted to a general type (read: interface)
# and then converted to the real type.

def _register_container_converters(converters: Iterable[Tuple[Tuple[type, ...], type]]) -> None:
    def _make_converter(cls: Callable, base_type: type):
        @functools.wraps(cls)
        def converter(val: Any):
            val = convert_value(val, base_type)
            if not isinstance(val, cls):
                val = cls(val)

            return val

        return converter

    for convs, base in converters:
        register_converter(*convs)(_make_converter(
            cast(Callable, convs[-1]),
            base,
        ))


_register_container_converters((
    ((Tuple, tuple), Iterable),
    ((List, list), Iterable),
    ((Set, set), Iterable),

    ((Dict, dict), Mapping),
))

del _register_container_converters


# <-- simple converters ------------------------------------------------------->

@register_converter(None, type(None))
def none_converter(_: Any) -> None:
    """Converts all values to `None`."""
    return None


@register_converter(Any)
def any_convert(val: Any) -> Any:
    """Returns value as is."""
    return val


@register_converter(Iterable, collections.Iterable)
def iterable_converter(value: Any) -> Iterable:
    """Converts the value to an iterable.

    Non-iterable values are wrapped in a tuple.

    Even though strings are iterable, this converter does not treat them as such
    to be consistent with the user's expectations.

    The same is true for mappings which is converted to an iterable of key,
    value tuples instead of just the keys.
    """
    if isinstance(value, Mapping):
        return value.items()

    if isinstance(value, Iterable) and not isinstance(value, (str,)):
        return value
    else:
        # yes it looks weird, but this is a tuple
        return value,


@register_converter(Mapping, collections.Mapping)
def mapping_converter(value: Any) -> Mapping:
    """Converts the value to a mapping,

    Sequences are interpreted as a mapping from index to value.
    All other value types raise a `ConversionError`.
    """
    if isinstance(value, Mapping):
        return value
    elif isinstance(value, Sequence):
        return dict(enumerate(value))
    else:
        raise ConversionError(f"can't convert {value!r} to a Mapping")


# <-- complex converters ------------------------------------------------------>

@register_converter()
class UnionConverter(ComplexConverterABC):
    """Converter for union types.

    First checks if value is already in the union and if it's not it then
    tries to convert to the values from first to last.
    """

    def can_convert(self, target: type) -> bool:
        return typeinspect.is_union(target) and all(map(has_converter, typeinspect.get_type_args(target)))

    def convert(self, value: Any, target: type) -> Any:
        if typeinspect.has_type(value, target):
            return value

        types = typeinspect.get_type_args(target)
        last_exception: Optional[Exception] = None

        for typ in types:
            try:
                return convert_value(value, typ)
            except ConversionError as e:
                if last_exception is not None:
                    last_exception.__cause__ = e

                last_exception = e

        raise last_exception


@register_converter()
class TupleConverter(ComplexConverterABC):
    """Converter for typed tuples.

    The input value is first converted to a collection, if the tuple
    type has a fixed length the input must match that, otherwise any
    length is accepted.
    """

    def can_convert(self, target: type) -> bool:
        is_tuple = typeinspect.is_tuple(target) \
                   and not typeinspect.has_free_parameters(target)

        if not is_tuple:
            return False

        container_type = typeinspect.get_origin(target)
        item_types = typeinspect.resolve_tuple(target)[0]
        return has_converter(container_type) and all(map(has_converter, item_types))

    def convert(self, value: Any, target: type) -> tuple:
        # convert to a collection
        values = convert_value(value, list)

        types, n = typeinspect.resolve_tuple(target)
        if n is None:
            typ = types[0]
            return tuple(convert_value(val, typ) for val in values)
        elif n != len(values):
            raise ConversionError(f"Can't convert {values!r} to {n}-tuple, lengths don't match")
        else:
            return tuple(convert_value(val, typ) for val, typ in zip(values, types))


@register_converter()
class IterableConverter(ComplexConverterABC):
    """Converts the value to a typed iterable.

    The value is first converted to an untyped `Iterable`.
    All items are converted to the item type and gathered in a `list`.
    The list is then converted to the container type.
    """

    def can_convert(self, target: type) -> bool:
        is_iterable = typeinspect.is_generic_iterable(target) \
                      and not typeinspect.has_free_parameters(target) \
                      and not typeinspect.is_tuple(target) \
                      and not typeinspect.is_generic_mapping(target)

        if not is_iterable:
            return False

        container_type = typeinspect.get_origin(target)
        item_type = typeinspect.get_type_args(target)[0]
        return has_converter(container_type) and has_converter(item_type)

    def convert(self, value: Any, target: type) -> Iterable[T]:
        container_type = typeinspect.get_origin(target)
        item_type = typeinspect.get_type_args(target)[0]

        final_list = []

        it = convert_value(value, Iterable)
        for i, sub_value in enumerate(it):
            try:
                v = convert_value(sub_value, item_type)
            except ConversionError as e:
                raise ConversionError(f"couldn't convert value at index {i} ({sub_value!r}) "
                                      f"to {typeinspect.friendly_name(item_type)}") from e

            final_list.append(v)

        if container_type is not None:
            final_list = convert_value(final_list, container_type, exclude_converters={self})

        return final_list


@register_converter()
class MappingConverter(ComplexConverterABC):
    """Converts the value to a mapping.

    The value is first converted to an untyped `Mapping`.
    A `dict` is which maps the keys converted to the key type to
    the values converted to the value type.
    This dict is then converted to the container type.
    """

    def can_convert(self, target: type) -> bool:
        is_mapping = typeinspect.is_generic_mapping(target) \
                     and not typeinspect.has_free_parameters(target)
        if not is_mapping:
            return False

        container_type = typeinspect.get_origin(target)
        key_type, value_type = typeinspect.get_type_args(target)

        return has_converter(container_type) \
               and has_converter(key_type) \
               and has_converter(value_type)

    def convert(self, value: Any, target: type) -> Iterable[T]:
        container_type = typeinspect.get_origin(target)
        key_type, value_type = typeinspect.get_type_args(target)

        final_map = {}

        mapping = convert_value(value, Mapping)
        for key, value in mapping.items():
            try:
                k = convert_value(key, key_type)
            except ConversionError as e:
                raise ConversionError(f"couldn't convert key {key!r} to "
                                      f"{typeinspect.friendly_name(key_type)}") from e

            try:
                v = convert_value(value, value_type)
            except ConversionError as e:
                raise ConversionError(f"couldn't convert value of {key!r} ({value!r}) "
                                      f"to {typeinspect.friendly_name(value_type)}") from e

            final_map[k] = v

        if container_type is not None:
            final_map = convert_value(final_map, container_type, exclude_converters={self})

        return final_map


@register_converter()
class TemplateConverter(ComplexConverterABC):
    """Converter which converts the value to a template-like object.

    This converter exists mainly for templates used in containers like
    `List[MyTemplate]` or `Dict[str, MyTemplate]`.

    It also requires that the value is a complete template object. This
    is why it shouldn't be used by a source.
    """

    def can_convert(self, target: type) -> bool:
        return templ.is_template_like(target)

    def convert(self, value: Any, target: type) -> Any:
        value_map = convert_value(value, Mapping)

        fields = templ.fields(target)
        obj = templ.create_object_from_template(target)
        source.load_fields_values(obj, fields, value_map)
        templ.ensure_complete(obj, target)

        return obj


@register_converter()
class EnumConverter(ComplexConverterABC):
    """Converter for converting values to `enum.Enum`.

    The converter prefers a perfect name match, if that fails it tries
    to use a perfect value match.
    If that also fails and the value is a string, the first case-insensitive
    match on either the name or the value of a filed is returned.
    """

    def can_convert(self, target: type) -> bool:
        return inspect.isclass(target) and issubclass(target, enum.Enum)

    def convert(self, value: Any, target: enum.EnumMeta) -> enum.Enum:
        try:
            return target[value]
        except KeyError:
            pass

        try:
            return target(value)
        except (ValueError, TypeError):
            pass

        if isinstance(value, str):
            value_lower = value.lower()
            for enum_field in target:
                enum_field = cast(enum.Enum, enum_field)

                if value_lower == enum_field.name.lower():
                    return enum_field

                field_val = enum_field.value
                if isinstance(field_val, str) and field_val.lower() == value_lower:
                    return enum_field

        raise ConversionError(f"{value!r} isn't in enum {typeinspect.friendly_name(target)}")



import pathlib
from typing import Any, Callable, Dict, Optional

import konfi

__all__ = ["FileLoaderType",
           "register_file_loader", "has_file_loader",
           "FileLoader"]

FileLoaderType = Callable[[str], konfi.SourceABC]
FileLoaderType.__doc__ = \
    """Constructor of a konfi file source.
    
    In addition to the path, the functions must also accept various keyword 
    arguments. Specifically, it should not raise an exception if unknown keyword
    arguments are given!
    """

_FILE_LOADERS: Dict[str, FileLoaderType] = {}


def register_file_loader(*file_types: str, replace: bool = False):
    """Decorator to register a file loader.

    The decorated object must be a `FileLoaderType`.

    Args:
        *file_types: File extensions (including the dot) to assign to
            the decorated source.
        replace: Whether to replace existing file loaders if an extension
            is already registered.

    Raises:
        TypeError: If the decorated value isn't a file loader constructor
        ValueError: If one of the file types is already registered and
            `replace` isn't `True`.
    """

    def decorator(loader: FileLoaderType):
        if not callable(loader):
            raise TypeError("decorated value must be a file loader constructor")

        for file_type in file_types:
            file_type = file_type.lower()
            if file_type in _FILE_LOADERS and not replace:
                raise ValueError(f"file loader for file type {file_type!r} already exists.")

            _FILE_LOADERS[file_type] = loader

        return loader

    return decorator


def has_file_loader(ext: str) -> bool:
    """Check whether the given extension has a file loader associated.

    Args:
        ext: File extension including the leading dot.
    """
    return ext in _FILE_LOADERS


class FileLoader(konfi.SourceABC):
    """A higher order source which uses other sources under the hood.

    The source uses the file extension of the given path to determine which
    source to use. This is done by providing the `register_file_loader`
    decorator which can be used to register a `FileLoaderType` (which can be a
    `konfi.SourceABC`) for the given extensions.

    Supported extensions by the built-in sources:

    - YAML: .yml, .yaml, .json
    - TOML: .toml

    The file loader is determined as soon as the constructor is called, so if
    there is no file loader for the given extension a `ValueError` is raised
    unless `ignore_no_loader` is set to `True`.

    Args:
        path: Path of config file to load.
        ignore_no_loader: If set to `True` and no loader could be found for
            the given path, the source turns into a dummy source and doesn't
            load anything.
        ignore_not_found: If set to `True` and the wrapped source raises
            a `FileNotFoundError`, it is ignored.
        **kwargs: Keyword arguments to pass to the file loader constructor.

    Raises:
        ValueError: If no file loader was found for the given path and
            `ignore_no_loader` is `False`.
    """
    _loader: Optional[konfi.SourceABC]

    _ignore_not_found: bool
    _kwargs: Dict[str, Any]

    def __init__(self, path: str, *,
                 ignore_no_loader: bool = False,
                 ignore_not_found: bool = False,
                 **kwargs: Any) -> None:
        suffix = pathlib.Path(path).suffix.lower()
        try:
            loader_cls = _FILE_LOADERS[suffix]
        except KeyError:
            if ignore_no_loader:
                loader = None
            else:
                raise ValueError(f"No loader for file type {suffix}")
        else:
            loader = loader_cls(path, **kwargs)

        self._loader = loader

        self._ignore_not_found = ignore_not_found

    def __str__(self) -> str:
        loader = self._loader or "no loader"
        return f"FileLoader ({loader})"

    def load_into(self, obj: Any, template: type) -> None:
        if self._loader is None:
            return

        try:
            self._loader.load_into(obj, template)
        except FileNotFoundError as e:
            if self._ignore_not_found:
                return
            else:
                raise e



"""konfi is a config parser."""

from .converter import ComplexConverterABC, ConversionError, ConverterABC, \
    ConverterFunc, ConverterType, convert_value, has_converter, \
    register_converter, unregister_converter
from .field import Field, MISSING, NoDefaultValue, UnboundField, ValueFactory, field
from .loader import Loader, SourceError
from .source import FieldError, MultiPathError, PathError, SourceABC
from .sources import *
from .templ import create_object_from_template, fields, get_field, is_template, is_template_like, template

# load built-in converters
from . import converters

__version__ = "0.2.1"
__author__ = "Giesela Inc."

default_loader = Loader()
set_sources = default_loader.set_sources
load = default_loader.load



from typing import Any

import konfi
from konfi.source import load_fields_values
from .file import register_file_loader

__all__ = ["TOML"]


@register_file_loader(".toml")
class TOML(konfi.SourceABC):
    _path: str
    _ignore_not_found: bool

    def __init__(self, path: str, *, ignore_not_found: bool = False, **_) -> None:
        self._path = path
        self._ignore_not_found = ignore_not_found

    def __str__(self) -> str:
        return f"TOML: {self._path!r}"

    def load_into(self, obj: Any, template: type) -> None:
        try:
            import toml
        except ImportError as e:
            raise ImportError("Couldn't import the 'toml' package. Make sure it's installed.") from e

        try:
            data = toml.load(self._path)
        except FileNotFoundError as e:
            if self._ignore_not_found:
                return
            else:
                raise e

        load_fields_values(obj, konfi.fields(template), data)



import abc
import functools
import inspect
import logging
from typing import Any, Callable, Generic, Iterable, List, MutableMapping, Optional, Set, Type, TypeVar, Union, \
    overload, Tuple

from . import typeinspect

__all__ = ["ConverterABC", "ConverterFunc", "ConverterType", "ComplexConverterABC",
           "is_converter_type", "is_complex_converter",
           "register_converter", "unregister_converter",
           "get_converters", "has_converter",
           "ConversionError", "convert_value"]

log = logging.getLogger(__name__)

CT = TypeVar("CT")


class ConverterABC(abc.ABC, Generic[CT]):
    """Converter which converts a value to a type.

    Usually a normal converter only converts to one specific type, or a group
    of closely related types.
    """

    @abc.abstractmethod
    def convert(self, value: Any, target: Type[CT]) -> CT:
        """Convert the given value to the target type.

        Args:
            value: Value to convert.
            target: Type to convert the value to.

        Raises:
            Exception: When the conversion failed.

        Returns:
            Converted value.
        """
        ...


ConverterFunc = Callable[[Any], CT]
ConverterFunc.__doc__ = \
    """Callable which converts the given value to a type."""

ConverterType = Union[ConverterABC, Type[ConverterABC], ConverterFunc]
ConverterType.__doc__ = \
    """Type of a converter.
    """


class ComplexConverterABC(ConverterABC[Any], abc.ABC):
    """

    Notes:
        Unlike `ConverterABC` converters, the complex converter will be converted
        to a singleton instance upon registration.
    """

    @abc.abstractmethod
    def can_convert(self, target: type) -> bool:
        ...


def is_converter_type(obj: Any) -> bool:
    """Check if the object is a converter (`ConverterType`)."""
    if isinstance(obj, ConverterABC) or inspect.isclass(obj) and issubclass(obj, ConverterABC):
        return True

    # we could check the signature too, but at some point we just have to
    # trust the user
    return callable(obj)


def is_complex_converter(obj: Any) -> bool:
    """Check if the object is a complex converter.`"""
    return isinstance(obj, ComplexConverterABC) or inspect.isclass(obj) and issubclass(obj, ComplexConverterABC)


_CONVERTERS: MutableMapping[Type, List[ConverterType]] = {}
_COMPLEX_CONVERTERS: List[ComplexConverterABC] = []


def register_converter(*types: Type):
    """Decorator which registers the underlying type as a converter.

    This can either be a converter function (`ConverterFunc`), a `ConverterABC`
    class or an instance thereof.
    If the converter is a class, it is instantiated only when the conversion
    takes place. The constructor will be called without any arguments.
    There is an exception for `ComplexConverterABC` classes, as they are
    instantiated upon registration.

    Args:
        *types: Types which the converter converts to.
            This is ignored if the converter is a complex converter, but
            required otherwise.

    Raises:
        TypeError: If the decorated object isn't a converter type.
        ValueError: If no types are provided but the converter
            isn't a complex converter or vice versa.
    """

    def decorator(target: ConverterType):
        if not is_converter_type(target):
            raise TypeError("decorated object must be a converter.")

        if is_complex_converter(target):
            if types:
                raise ValueError("no types can be given for complex converters, use the can_convert method instead.")

            if inspect.isclass(target):
                try:
                    target_inst = target()
                except Exception as e:
                    raise ValueError(f"Couldn't create instance of complex converter {target!r}") from e
            else:
                target_inst = target

            _COMPLEX_CONVERTERS.append(target_inst)
        else:
            if not types:
                raise ValueError("at least one target type must be provided for non-complex converters")

            for typ in types:
                cs = _CONVERTERS.setdefault(typ, [])
                cs.append(target)

        _clear_converter_cache()
        return target

    return decorator


def unregister_converter(conv: ConverterType, *types: type) -> None:
    """Unregister the given converter from the given types.

    Args:
        conv: Converter to unregister
        *types: Types to unregister converter from. If empty, the converter
            is unregistered from all types.

    Raises:
        TypeError: If the given converter is a complex converter and types
            are given.
    """
    if is_complex_converter(conv):
        if types:
            raise TypeError("can't unregister a complex converter from specific types")

        if inspect.isclass(conv):
            for _conv in reversed(_COMPLEX_CONVERTERS):
                if type(_conv) is conv:
                    conv = _conv
                    break
            else:
                return

        try:
            _COMPLEX_CONVERTERS.remove(conv)
        except ValueError:
            return
    else:
        if not types:
            types = _CONVERTERS.keys()

        for typ in types:
            try:
                converters = _CONVERTERS[typ]
            except KeyError:
                continue

            try:
                converters.remove(conv)
            except ValueError:
                pass

    _clear_converter_cache()


@functools.lru_cache(maxsize=None)
def _get_complex_converters(target: Type) -> Tuple[ComplexConverterABC]:
    return tuple(c for c in reversed(_COMPLEX_CONVERTERS) if c.can_convert(target))


def get_converters(target: Type) -> Iterable[ConverterType]:
    """Get the converters for the given target type.

    Complex converters are only returned if no other converter could be found,
    The converters are ordered in the reverse order of their registration.
    """
    # iterate in reverse because we want custom converters to override built-ins
    try:
        return tuple(reversed(_CONVERTERS[target]))
    except KeyError:
        return _get_complex_converters(target)


def _clear_converter_cache() -> None:
    _get_complex_converters.cache_clear()


def has_converter(target: Type) -> bool:
    """Check whether the given target type has a converter."""
    try:
        next(iter(get_converters(target)))
    except StopIteration:
        return False
    else:
        return True


class ConversionError(Exception):
    """Exception raised if a conversion failed."""
    ...


def _call_converter(conv: ConverterType, value: Any, target: Type) -> Any:
    if inspect.isclass(conv) and issubclass(conv, ConverterABC):
        try:
            conv = conv()
        except Exception as e:
            raise ConversionError(f"Couldn't create an instance of the converter {conv}") from e

    if isinstance(conv, ConverterABC):
        converter_func = functools.partial(conv.convert, value, target)
    elif callable(conv):
        converter_func = functools.partial(conv, value)
    else:
        raise TypeError(f"Converter must be a converter type, not {conv!r}")

    try:
        return converter_func()
    except ConversionError as e:
        raise e
    except Exception as e:
        raise ConversionError(f"Couldn't convert {value!r} to "
                              f"{typeinspect.friendly_name(target)} using {conv}") from e


T = TypeVar("T")


@overload
def convert_value(value: Any, target: Type[T], *, exclude_converters: Set[ConverterType] = None) -> T: ...


@overload
def convert_value(value: Any, target: type, *, exclude_converters: Set[ConverterType] = None) -> Any: ...


def convert_value(value: Any, target: Type[T], *, exclude_converters: Set[ConverterType] = None) -> T:
    """Convert the value to the given type.

    If no converter was found but the target type is a class,
    a conversion using the constructor is attempted.

    If the value already has the target type, it is returned even if no
    converters are found.

    Args:
        value: Value to convert
        target: Target type to convert to.

        exclude_converters: Set of converter types to exclude.
            Note that this only excludes the converters from this call,
            if a converter further down the line calls this function the
            exclusion no longer applies.

    Raises:
        ConversionError: If the conversion failed.
    """
    converters = get_converters(target)

    last_exception: Optional[Exception] = None
    for c in converters:
        if exclude_converters and c in exclude_converters:
            continue

        try:
            converted = _call_converter(c, value, target)
        except ConversionError as e:
            if last_exception is not None:
                e.__cause__ = last_exception

            last_exception = e
        else:
            return converted

    # if we already have the correct type then just move on
    if typeinspect.has_type(value, target):
        return value

    # no converter found
    if last_exception is None:
        if inspect.isclass(target):
            log.debug("no converters found, trying constructor")
            try:
                return target(value)
            except Exception as e:
                raise ConversionError(f"No converter found and couldn't use {target!r} directly") from e

        raise ConversionError(f"No converter found for {target!r}")

    raise last_exception



import ast
import os
import re
from typing import Any, Callable, Dict, Iterable, List, Pattern, Type, Union

import konfi
from konfi import source

__all__ = ["Env",
           "Decoder", "ResolvableDecoder", "resolve_decoder",
           "NameBuilder", "build_env_name"]

unportable_chars_pattern: Pattern = re.compile(r"^\s*\d+|[^a-z0-9]+", re.IGNORECASE)

Decoder = Callable[[str], Any]
Decoder.__doc__ = \
    """Callable which decodes the value of an environment variable."""

_DECODERS: Dict[str, Union[Decoder, Exception]] = {
    "raw": str,
    "python": ast.literal_eval,
}

try:
    import yaml
except ImportError as e:
    exc = ImportError("yaml decoder unavailable because pyyaml package is missing.")
    exc.__cause__ = e
    _DECODERS["yaml"] = e
else:
    _DECODERS["yaml"] = yaml.safe_load

ResolvableDecoder = Union[str, Decoder]
ResolvableDecoder.__doc__ = \
    """Either the name of a built-in decoder or a `Decoder` function."""


def resolve_decoder(decoder: ResolvableDecoder) -> Decoder:
    """Resolve the given decoder name or check if it's a decoder function.

    Args:
        decoder: Decoder to resolve

    Raises:
        KeyError: Decoder name was given, but doesn't exist
        TypeError: Invalid type was given
        Exception: If the decoder whose name was given is unavailable.
            For example if the `pyyaml` package isn't installed, an
            `ImportError` will be raised when trying to resolve the
            "yaml" decoder.
    """
    if isinstance(decoder, str):
        try:
            dec = _DECODERS[decoder.lower()]
        except KeyError:
            raise KeyError(f"No existing decoder {decoder!r}")
        else:
            if isinstance(dec, Exception):
                raise dec from None
            else:
                return dec

    if callable(decoder):
        return decoder
    else:
        raise TypeError(f"Decoder must be a function or the name of a built-in decoder")


NameBuilder = Callable[[List[str]], str]
NameBuilder.__doc__ = \
    """A callable which converts a path to the name of the corresponding 
    environment variable."""


def build_env_name(path: Iterable[str]) -> str:
    """Create the name of an environment variable for the given path."""
    return "_".join(unportable_chars_pattern.sub("", part) for part in path).upper()


class Env(konfi.SourceABC):
    """Source which loads the config from environment variables.

    The env source is different from most sources because it walks through the
    config template and looks for the corresponding environment variable instead
    of the other way around. This means that with the exception of template-like
    objects it's not possible to set sub values directly. For example, you can't
    update a specific key of a dictionary using an environment variable using a
    specific environment variable, only the entire value (i.e. dictionary) can
    be set.


    Args:
          prefix: Prefix to prepend to all variable names.
            This can be used to prevent name collisions and ensure
            that the variables were set with the right intent.

          decoder: Decoder used to interpret the values of the environment
              variables. There are three built-in decoders:

              - raw: Values are interpreted as strings.

              - python: Values are (safely) interpreted as if they were Python
                literals.

              - yaml: Values are interpreted as YAML. This is by far the most
                powerful and convenient decoder. For example, it doesn't require
                the use of quotation marks to escape strings.


              You can also pass your own decoder with the type `Decoder` which
              is just a function that takes a string and returns the decoded
              version.

              The default is the python decoder.

          name_builder: Function that combines the path segments of a field to
            the name of the corresponding environment variable
            (ex: ["database", "main", "url"] -> "DATABASE_MAIN_URL").

            The default is the `build_env_name` function which removes
            non-alphanumeric characters and underscores from the path segments
            as well as stripping leading numbers. The segments are then joined
            using underscores. For example ["a_b@", "1c_d"] would be turned
            into "AB_CD".
    """
    _prefix: str
    _decoder: Decoder
    _name_builder: NameBuilder

    def __init__(self, prefix: str = "", *,
                 decoder: ResolvableDecoder = "python",
                 name_builder: NameBuilder = build_env_name,
                 ):
        self._prefix = prefix
        self._decoder = resolve_decoder(decoder)
        self._name_builder = name_builder

    def __str__(self) -> str:
        return f"Environment ({self._prefix!r})"

    def get_env_name(self, path: List[str]) -> str:
        """Combine prefix and name builder result."""
        return f"{self._prefix}{self._name_builder(path)}"

    def _load_path(self, path: List[str], parent: Any, field: konfi.Field) -> None:
        key = self.get_env_name(path)
        try:
            raw_value = os.environ[key]
        except KeyError:
            return

        try:
            value = self._decoder(raw_value)
        except Exception as e:
            raise konfi.PathError(path, f"can't decode {raw_value!r}") from e

        source.load_field_value(parent, field, value)

    def load_into(self, obj: Any, template: Type) -> None:
        for qfield in source.iter_fields_recursively(obj, template):
            self._load_path(qfield.path, qfield.parent, qfield.field)



import inspect
from typing import List, Type, TypeVar, Union

from .source import SourceABC
from .templ import create_object_from_template, ensure_complete, is_template_like

__all__ = ["SourceError", "Loader"]

TT = TypeVar("TT")


class SourceError(Exception):
    """Exception raised when a source fails to load the config.

    Attributes:
        source (SourceABC): Source that tried to load the config.
        template (type): Template that was to be loaded.
        error (Exception): Original error that was raised.
            This is provided for convenience.
    """
    source: SourceABC
    template: type

    error: Exception

    def __init__(self, source: SourceABC, template: type, error: Exception) -> None:
        self.source = source
        self.template = template
        self.error = error

    def __str__(self) -> str:
        return f"couldn't load {self.template.__qualname__!r} from {self.source}"


class Loader:
    """Loads the config from sources."""
    _sources: List[SourceABC]

    def __init__(self) -> None:
        self._sources = []

    def set_sources(self, *sources: SourceABC) -> None:
        """Set the sources to use when loading."""
        self._sources = list(sources)

    def load(self, template: Union[Type[TT], TT]) -> TT:
        """Load the config for the given template.

        Args:
            template: Template to load from the sources.
                If this is already an instance of the template, then
                the config is loaded into the instance.

        Raises:
            ValueError: If no sources are set.
            TypeError: If the given template isn't template-like.
            SourceError: If one of the sources fails to load the config.
        """

        if not self._sources:
            raise ValueError("No sources are set")

        if not is_template_like(template):
            raise TypeError(f"Template must be template-like, not {template!r}")

        if inspect.isclass(template):
            obj = create_object_from_template(template)
        else:
            obj = template
            template = type(template)

        for source in self._sources:
            try:
                source.load_into(obj, template)
            except Exception as e:
                raise SourceError(source, template, e) from e

        ensure_complete(obj, template)

        # TODO add hooks to the entire process

        return obj



"""Runtime introspection for typings."""

import inspect
import sys
from typing import Any, Generic, Iterable, Mapping, Optional, Tuple, TypeVar, \
    Union, _GenericAlias, _eval_type, Dict, ForwardRef

TypeTuple = Tuple[type, ...]
TypeTuple.__doc__ = \
    """Variadic tuple containing types."""

NoneType = type(None)


def class_and_subclass(typ: type, cls: type) -> bool:
    return inspect.isclass(typ) and issubclass(typ, cls)


def get_origin(typ: type) -> Optional[type]:
    if isinstance(typ, _GenericAlias):
        return typ.__origin__

    return None


def has_origin(typ: type, origin: type) -> bool:
    typ_origin = get_origin(typ)
    if typ_origin is None:
        return False

    return class_and_subclass(typ_origin, origin)


def get_type_args(typ: type) -> TypeTuple:
    if isinstance(typ, _GenericAlias):
        return typ.__args__

    return ()


def get_parameters(typ: type) -> TypeTuple:
    if isinstance(typ, _GenericAlias):
        return typ.__parameters__

    return ()


def has_free_parameters(typ: type) -> bool:
    return len(get_parameters(typ)) > 0


def is_any(typ: type) -> bool:
    # TODO typevar?
    return typ is Any


def is_union(typ: type) -> bool:
    return get_origin(typ) is Union


def is_tuple(typ: type) -> bool:
    return typ is Tuple \
           or get_origin(typ) is tuple \
           or is_generic(typ) and class_and_subclass(typ, tuple)


def is_typevar(typ: type) -> bool:
    return isinstance(typ, TypeVar)


def is_generic(typ: type) -> bool:
    return class_and_subclass(typ, Generic) \
           or isinstance(typ, _GenericAlias)


def is_generic_iterable(typ: type) -> bool:
    return is_generic(typ) and has_origin(typ, Iterable)


def is_generic_mapping(typ: type) -> bool:
    return is_generic(typ) and has_origin(typ, Mapping)


def is_optional(typ: type) -> bool:
    if typ is NoneType:
        return True
    elif is_union(typ):
        return any(is_optional(t) for t in get_type_args(typ))
    # TODO typevar ?
    else:
        return False


def resolve_tuple(typ: type) -> Tuple[TypeTuple, Optional[int]]:
    args = get_type_args(typ)
    # homogeneous variadic tuple
    if len(args) == 2 and args[1] == Ellipsis:
        return (args[0],), None
    else:
        return args, len(args)


def _has_union_type(obj: Any, union: type) -> bool:
    return any(has_type(obj, typ) for typ in get_type_args(union))


def _has_tuple_type(obj: Any, tup: type) -> bool:
    if not isinstance(obj, tuple):
        return False

    types, n = resolve_tuple(tup)
    if n is None:
        typ = types[0]
        return all(has_type(val, typ) for val in obj)
    elif n != len(obj):
        return False
    else:
        return all(has_type(val, typ) for val, typ in zip(obj, types))


def _has_generic_iterable_type(obj: Any, iter_type: type) -> bool:
    container_type = get_origin(iter_type)
    if not has_type(obj, container_type):
        return False

    item_type = get_type_args(iter_type)[0]
    return all(has_type(item, item_type) for item in obj)


def _has_generic_mapping_type(obj: Any, map_type: type) -> bool:
    container_type = get_origin(map_type)
    if not has_type(obj, container_type):
        return False

    key_type, value_type = get_type_args(map_type)
    return all(has_type(key, key_type) and has_type(value, value_type) for key, value in obj.items())


def has_type(obj: Any, typ: type) -> bool:
    if is_any(typ):
        return True

    if is_union(typ):
        return _has_union_type(obj, typ)

    if is_tuple(typ):
        return _has_tuple_type(obj, typ)

    if is_generic_iterable(typ):
        return _has_generic_iterable_type(obj, typ)

    if is_generic_mapping(typ):
        return _has_generic_mapping_type(obj, typ)

    # TODO this doesn't feel safe
    return isinstance(obj, typ)


def friendly_name(typ: Any) -> str:
    if inspect.isclass(typ) and typ.__module__ == "builtins":
        return typ.__qualname__
    else:
        if is_union(typ):
            args = get_type_args(typ)
            if len(args) == 2 and args[1] is NoneType:
                return f"typing.Optional[{friendly_name(args[0])}]"

        return repr(typ)


def get_class_type_hints_strict(cls: type,
                                globalns: Mapping[str, Any] = None,
                                localns: Mapping[str, Any] = None
                                ) -> Dict[str, type]:
    if not inspect.isclass(cls):
        raise TypeError("cls needs to be a class. Use typing.get_type_hints for other types")

    if globalns is None:
        globalns = sys.modules[cls.__module__].__dict__

    ann = cls.__dict__.get("__annotations__", {})

    hints = {}

    for name, value in ann.items():
        if value is None:
            value = type(None)
        if isinstance(value, str):
            value = ForwardRef(value, is_argument=False)

        value = _eval_type(value, globalns, localns)
        hints[name] = value

    return hints



import setuptools

import konfi

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="konfi",
    version=konfi.__version__,
    author=konfi.__author__,
    author_email="team@giesela.dev",
    description="Python config management, but cute.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gieseladev/konfi",
    license="MIT",

    packages=setuptools.find_packages(exclude=("examples", "tests")),

    # entry_points={
    #     "console_scripts": [
    #         "konfi=konfi.cli:main",
    #     ]
    # },

    install_requires=["toml", "pyyaml"],

    extras_require={
        "docs": ["sphinx", "sphinx-autodoc-typehints"],
    },
)



from typing import Any, Callable, Optional, TYPE_CHECKING, Type, TypeVar

from . import typeinspect
from .converter import ConverterType

__all__ = ["ValueFactory", "MISSING",
           "NoDefaultValue",
           "UnboundField", "Field",
           "upgrade_unbound", "field"]

FT = TypeVar("FT")
ValueFactory = Callable[[], Any]
ValueFactory.__doc__ = \
    """Callable which when called, returns a value."""

# only use "expensive" sentinel for documentation
if TYPE_CHECKING:
    class MISSING:
        """Sentinel representing a missing value.

        This is used to represent the lack of a default value, because
        `None` would lead to a conflict.
        """

        def __repr__(self) -> str:
            return "MISSING"


    MISSING = MISSING()
else:
    MISSING = object()


class NoDefaultValue(Exception):
    """Exception for when a default value is expected, but doesn't exist."""
    ...


class UnboundField:
    """A field that hasn't been bound to a template.

    Note that an unbound field is usually created using the `field`
    function, not directly.

    Attributes:
        key (Optional[str]): Corresponding config key to use.
        comment (Optional[str]): Comment for the field.
        default_value (Any): The default value of the field.
            The value `MISSING` is used if no default value exists.
        default_factory (Optional[ValueFactory]): Factory to call to
            get the default value.
        converter (Optional[ConverterType]): Converter to use.

    Raises:
        ValueError: If both factory and default value are specified.
    """
    key: Optional[str]
    comment: Optional[str]

    default_value: Any
    default_factory: Optional[ValueFactory]

    converter: Optional[ConverterType]

    def __init__(self, *,
                 key: str = None,
                 comment: str = None,
                 default_value: Any = MISSING,
                 default_factory: ValueFactory = None,
                 converter: ConverterType = None,
                 ) -> None:
        self.key = key
        self.comment = comment

        if default_value is not MISSING and default_factory is not None:
            raise ValueError("can't specify both default factory and value!")

        self.default_value = default_value
        self.default_factory = default_factory

        self.converter = converter

    def __repr__(self) -> str:
        return f"UnboundField(key={self.key!r}, comment={self.comment!r})"

    @property
    def required(self) -> bool:
        """Checks whether the field must be set."""
        return self.default_value is MISSING and self.default_factory is None

    def get_default(self) -> Any:
        """Get the default value of the field.

        This uses either the default value or calls the default factory.

        Raises:
            NoDefaultValue: If the field doesn't have a default value.
        """
        if self.default_value is not MISSING:
            return self.default_value
        elif self.default_factory is not None:
            return self.default_factory()
        else:
            raise NoDefaultValue


class Field(UnboundField):
    """A field of a template.

    Field is a superset of `UnboundField` meaning it inherits all attributes
    and arguments.

    Attributes:
        attribute (str): Name of the attribute the field belongs to.
        key (str): Name of the config key.
        value_type (Type): Expected type of the field.
    """
    attribute: str
    key: str

    value_type: Type

    def __init__(self, *,
                 attribute: str,
                 key: str = None,
                 comment: str = None,
                 value_type: Type,
                 default_value: Any = MISSING,
                 default_factory: ValueFactory = None,
                 converter: ConverterType = None,
                 ) -> None:
        if key is None:
            key = attribute

        super().__init__(key=key, comment=comment,
                         default_value=default_value, default_factory=default_factory,
                         converter=converter)
        self.attribute = attribute
        self.value_type = value_type

    def __repr__(self) -> str:
        return f"Field(attribute={self.attribute!r}, value_type={self.value_type!r}, key={self.key!r}, comment={self.comment!r})"

    def __str__(self) -> str:
        key_str = f" [{self.key!r}]" if self.key != self.attribute else ""
        return f"Field({self.attribute}{key_str})"

    @property
    def optional_type(self) -> bool:
        """Whether the value type is optional (ex: Optional[str])."""
        return typeinspect.is_optional(self.value_type)

    @property
    def required(self) -> bool:
        return super().required and not self.optional_type

    def get_default(self) -> Any:
        try:
            return super().get_default()
        except NoDefaultValue as e:
            if self.optional_type:
                return None
            else:
                raise e


def upgrade_unbound(unbound: UnboundField, *, attribute: str, value_type: Type) -> Field:
    """Upgrade an unbound field to a field.

    Args:
        unbound: Unbound field to upgrade.
        attribute: Attribute name of the field.
        value_type: Value type of the field.

    Returns:
        A field with the values of the unbound field.
    """
    return Field(
        attribute=attribute,
        key=unbound.key,
        comment=unbound.comment,
        value_type=value_type,
        default_value=unbound.default_value,
        default_factory=unbound.default_factory,
        converter=unbound.converter,
    )


def field(*, key: str = None, comment: str = None,
          default: Any = MISSING, factory: ValueFactory = None,
          converter: ConverterType = None,
          ) -> UnboundField:
    """Specify a field options.

    This creates an unbound field which is later upgraded to a bound field when
    the template is created. The class variable is also replaced with the
    default value of the field and if no default value exists, it is removed.

    Args:
        key: Config key to use (instead of the attribute name).
        comment: Comment for the field.
        default: Default value for the field.
        factory: Factory method to use to get the default value.
            You can't set both the default and the factory value.
        converter: Custom converter to use.

    Raises:
        ValueError: If both factory and default value are specified.

    Returns:
        An unbound field.
    """
    return UnboundField(
        key=key,
        comment=comment,
        default_value=default,
        default_factory=factory,
        converter=converter,
    )



import dataclasses
import functools
import inspect
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, get_type_hints, overload

from . import typeinspect
from .converter import has_converter
from .field import Field, MISSING, NoDefaultValue, UnboundField, upgrade_unbound
from .source import FieldError, MultiPathError, PathError

__all__ = ["template", "is_template", "is_template_like",
           "fields", "get_field",
           "create_object_from_template",
           "ensure_complete"]

_FIELDS_ATTR = "__template_fields__"


def _make_fields(cls: type, *, template_bases_only: bool = True) -> Dict[str, Field]:
    cls_fields: Dict[str, Field] = {}

    if template_bases_only:
        for base in cls.__mro__[:0:-1]:
            try:
                _fields = getattr(base, _FIELDS_ATTR)
            except AttributeError:
                continue

            for attr, _field in _fields.items():
                cls_fields[attr] = _field

        cls_annotations = typeinspect.get_class_type_hints_strict(cls)
    else:
        cls_annotations = get_type_hints(cls)

    for attr, typ in cls_annotations.items():
        try:
            value = getattr(cls, attr)
        except AttributeError:
            cls_fields[attr] = Field(attribute=attr, value_type=typ)
            continue

        if isinstance(value, UnboundField):
            field = upgrade_unbound(value, attribute=attr, value_type=typ)

            # replace field with the default if possible, otherwise remove
            if value.default_value is MISSING:
                delattr(cls, attr)
            else:
                setattr(cls, attr, value.default_value)

        else:
            field = Field(attribute=attr, value_type=typ, default_value=value)

        if not (field.converter is not None
                or is_template_like(typ)
                or has_converter(typ)):
            raise TypeError(f"Field {attr!r} doesn't have a converter.")

        cls_fields[attr] = field

    return cls_fields


def _make_template(cls: type, *, template_bases_only: bool = True):
    _fields = _make_fields(cls, template_bases_only=template_bases_only)
    setattr(cls, _FIELDS_ATTR, _fields)


def template(*, template_bases_only: bool = True):
    """Decorator to convert the given class to a template.

    This parses the decorated class into a template which can be used to
    load the config.
    The fields are loaded from the annotations from the class. If an attribute
    has a value it is used as the default value, unless it's an unbound field
    (see `konfi.field`) which is used directly and the class variable is
    replaced.

    Args:
        template_bases_only: If set to `True` (default) only fields from
            bases which are themselves templates are added to the template.
            If `False`, all annotations from all bases are considered as being
            part of the class itself.

    Raises:
        ValueError: If the decorated object isn't a class.
        TypeError: If one of the fields has an invalid type.
    """

    def decorator(cls: type):
        if not inspect.isclass(cls):
            raise ValueError("decorator must be applied to a class")

        _make_template(cls, template_bases_only=template_bases_only)
        return cls

    return decorator


def is_template(obj: Any) -> bool:
    """Check whether the given object is a template instance or class."""
    return hasattr(obj, _FIELDS_ATTR)


def is_template_like(obj: Any) -> bool:
    """Check whether the given object is template-like.

    Currently this includes templates and dataclasses.
    """
    return is_template(obj) or dataclasses.is_dataclass(obj)


def _field_from_dataclass_field(f: dataclasses.Field) -> Field:
    default_value = f.default
    if default_value is dataclasses.MISSING:
        default_value = MISSING

    default_factory = f.default_factory
    if default_factory is dataclasses.MISSING:
        default_factory = None

    return Field(attribute=f.name, key=f.name,
                 value_type=f.type,
                 default_value=default_value, default_factory=default_factory)


@functools.lru_cache(64)
def _fields_from_dataclass(obj: Any) -> Dict[str, Field]:
    data_fields: Tuple[dataclasses.Field, ...] = dataclasses.fields(obj)
    return {f.name: _field_from_dataclass_field(f) for f in data_fields}


def _get_fields(obj: Any) -> Dict[str, Field]:
    try:
        return getattr(obj, _FIELDS_ATTR)
    except AttributeError:
        if dataclasses.is_dataclass(obj):
            return _fields_from_dataclass(obj)

        raise TypeError("must be called with a template")


def fields(obj: Any) -> Tuple[Field, ...]:
    """Get the fields of a template-like instance or class.

    Raises:
        TypeError: If the given object isn't template-like.

    Returns:
        A tuple containing all fields of the template.
    """
    return tuple(_get_fields(obj).values())


def get_field(obj: Any, attr: str) -> Optional[Field]:
    """Get the field for an attribute from a template-like  instance or class.

    Raises:
        TypeError: If the given object isn't template-like.

    Returns:
        The field for the attribute or `None` if no such field exists.
    """
    return _get_fields(obj).get(attr)


T = TypeVar("T")


@overload
def create_object_from_template(templ: type) -> Any: ...


@overload
def create_object_from_template(templ: Type[T]) -> T: ...


def create_object_from_template(templ: Type[T]) -> T:
    """Create an empty object for the given template.

    Args:
        templ: Template-like object to create an instance for.

    Returns:
        New object which is an instance of the template.
    """
    return object.__new__(templ)


def ensure_complete(obj: Any, templ: type) -> None:
    """Check whether the given object contains all fields of the template.

    Sub-templates are checked recursively. If a field doesn't have a value but
    the field has a default value, the default value is assigned.
    No type checking is performed.

    Raises:
        TypeError: If the given template isn't template-like.
        PathError: If there's an error with one of the fields.
            `FieldError` is raised when a value is missing.
            If there are multiple errors, a `MultiPathError` exception
            is raised.
    """
    path_errors: List[PathError] = []

    for f in fields(templ):
        try:
            val = getattr(obj, f.attribute)
        except AttributeError:
            try:
                default_val = f.get_default()
            except NoDefaultValue:
                e = FieldError([f.key], f, "required value missing")
                path_errors.append(e)
            else:
                setattr(obj, f.attribute, default_val)

            continue

        if is_template_like(f.value_type):
            try:
                ensure_complete(val, f.value_type)
            except PathError as e:
                e.backtrace_path(f.key)
                path_errors.append(e)

    if len(path_errors) == 1:
        raise path_errors[0]
    elif path_errors:
        raise MultiPathError([], path_errors, f"{templ.__qualname__!r} is incomplete")



import abc
import dataclasses
import functools
from typing import Any, Dict, Iterable, Iterator, List, Mapping

import konfi

__all__ = ["SourceABC",
           "PathError", "MultiPathError",
           "FieldError",
           "QualifiedField", "iter_fields_recursively",
           "load_field_value", "load_fields_values"]


class SourceABC(abc.ABC):
    """Abstract base class of a source which can load the config."""

    @abc.abstractmethod
    def load_into(self, obj: Any, template: type) -> None:
        """Load the config into the given object according to the template.

        Raises:
            Exception: If the config couldn't be loaded.
        """
        ...


class PathError(Exception):
    """General exception in a template path.

    The path represents the config keys, which is not necessarily the same
    as the attributes.

    Attributes:
        path (List[str]): Path to the origin of the exception.
    """
    path: List[str]

    def __init__(self, path: Iterable[str], msg: str) -> None:
        super().__init__(msg)
        self.path = list(path)

    def __str__(self) -> str:
        path_str = self.path_str
        if path_str:
            path_str = f"{path_str}: "

        return f"{path_str}{super().__str__()}"

    @property
    def path_str(self) -> str:
        """Get the path as a string."""
        return ".".join(f"{path!r}" for path in self.path)

    def backtrace_path(self, part: str) -> None:
        """Add a part to the path."""
        self.path.insert(0, part)


class MultiPathError(PathError):
    """Exception grouping multiple `PathError` instances.

    Attributes:
        errors (List[PathError]): Grouped path errors.
    """
    errors: List[PathError]

    def __init__(self, path: Iterable[str], errors: Iterable[PathError], msg: str) -> None:
        super().__init__(path, msg)
        self.errors = list(errors)

    def __iter__(self) -> Iterable[PathError]:
        return iter(self.errors)

    def __str__(self) -> str:
        import textwrap
        errors_str = textwrap.indent("\n".join(map(str, self.errors)), "  ")
        return f"{super().__str__()}:\n{errors_str}"


class FieldError(PathError):
    """Exception for a particular field.

    Attributes:
        field (konfi.Field): Field which caused the exception.
    """
    field: konfi.Field

    def __init__(self, path: Iterable[str], field: konfi.Field, msg: str) -> None:
        super().__init__(path, msg)
        self.field = field


@dataclasses.dataclass()
class QualifiedField:
    """Qualified field information as returned by `iter_fields_recursively`."""
    parent: Any
    path: List[str]
    field: konfi.Field


def iter_fields_recursively(obj: Any, template: Any) -> Iterator[QualifiedField]:
    """Iterate over all fields of the template recursively.

     Yields:
          `QualifiedField` instances.
     """

    def iter_fields(parent: Any, path: List[str], fields: Iterable[konfi.Field]) -> None:
        for field in fields:
            key_path = [*path, field.key]

            if konfi.is_template_like(field.value_type):
                sub_obj = _get_sub_obj(obj, field)
                yield from iter_fields(sub_obj, key_path, konfi.fields(field.value_type))
            else:
                yield QualifiedField(parent, key_path, field)

    yield from iter_fields(obj, [], konfi.fields(template))


def load_field_value(obj: Any, field: konfi.Field, value: Any) -> None:
    """Load the given value for a field into the object.

    Raises:
        FieldError: If the value couldn't be converted to the given field
    """
    try:
        if field.converter is None:
            converted = konfi.convert_value(value, field.value_type)
        else:
            converted = konfi.converter._call_converter(field.converter, value, field.value_type)
    except konfi.ConversionError as e:
        raise FieldError([field.key], field, str(e)) from e

    setattr(obj, field.attribute, converted)


def _get_sub_obj(obj: Any, field: konfi.Field) -> Any:
    try:
        value = getattr(obj, field.attribute)
    except AttributeError:
        value = konfi.create_object_from_template(field.value_type)
        setattr(obj, field.attribute, value)

    return value


def load_fields_values(obj: Any, fields: Iterable[konfi.Field], mapping: Mapping, *,
                       ignore_unknown: bool = False) -> None:
    """Load the values for the fields from the mapping.

    This is done recursively.

    Args:
        obj: Object to load fields into
        fields: Fields to load
        mapping: Mapping to get field values from
        ignore_unknown: If `True`, excessive keys in the mapping are ignored.

    Raises:
        PathError: If there is an issue with a field.
            If multiple fields have an error, `MultiPathError` is raised.
    """
    _field_by_keys: Dict[str, konfi.Field] = {field.key: field for field in fields}

    field_errors: List[PathError] = []

    for key, value in mapping.items():
        try:
            field = _field_by_keys[key]
        except KeyError:
            if ignore_unknown:
                continue
            else:
                raise PathError([key], f"unexpected config key: {key!r}")

        if konfi.is_template_like(field.value_type) and isinstance(value, Mapping):
            sub_obj = _get_sub_obj(obj, field)
            loader = functools.partial(load_fields_values, sub_obj, konfi.fields(sub_obj), value)
        else:
            loader = functools.partial(load_field_value, obj, field, value)

        try:
            loader()
        except PathError as e:
            field_errors.append(e)
        except Exception as e:
            err = FieldError([key], field, str(e))
            err.__cause__ = e
            field_errors.append(err)

    if len(field_errors) == 1:
        raise field_errors[0]
    elif field_errors:
        raise MultiPathError((), field_errors, "")
